/*
Copyright (c) 2006 by Juliusz Chroboczek

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
*/

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <netinet/in.h>
#include <endian.h>
#include <byteswap.h>
#include <time.h>

#include "ahcp-protocol.h"

#define MAXSIZE 2048

unsigned int serial = 2006120304;
unsigned int ttl_delta = 600;
char *network_name = NULL;
unsigned char ipv6_prefix[16] =
    {0x20, 0x01, 0x06, 0x60, 0x33, 0x01, 0x80, 0x63,
     0, 0, 0, 0, 0, 0, 0, 0};
unsigned char routing_protocol = ROUTING_PROTOCOL_OLSR;
unsigned char *default_gateway = NULL;
unsigned char name_server[16] =
    {0x20, 0x01, 0x06, 0x60, 0x33, 0x01, 0x80, 0x63,
     0, 0, 0, 0, 0, 0, 0, 1};
unsigned char multicast_address[16] =
    { 0xff, 0x05, 0, 0, 0, 0, 0, 0,
      0xcc, 0xa6, 0xc0, 0xf9, 0xe1, 0x82, 0x53, 0x59 };
unsigned char toto[4] = "toto";

static unsigned long long
htonll(unsigned long long value)
{
#if BYTE_ORDER == LITTLE_ENDIAN
    return bswap_64(value);
#else
    return value;
#endif
}

int
main(int argc, char **argv)
{
    unsigned char buf[MAXSIZE];
    union {
        unsigned short s;
        char c[2];
    } twobytes;
    union {
        unsigned int i;
        char c[4];
    } fourbytes;
    union {
        unsigned long long i;
        char c[8];
    } eightbytes;
    int i;
    unsigned int ttl;
    char *usage = "ahcp-generate [-s serial] [-t ttl] [-n name] > ahcp.dat";
    int data_length;
    
    while(1) {
        int c;
        c = getopt(argc, argv, "s:t:n:");
        if(c < 0) break;
        switch(c) {
        case 'n': network_name = optarg; break;
        case 's': serial = atoi(optarg); break;
        case 't': ttl_delta = atoi(optarg); break;
        default: fprintf(stderr, "%s\n", usage); exit(1); break;
        }
    }

    if(argc != optind) {
        fprintf(stderr, "%s\n", usage); exit(1);
    }

    ttl = time(NULL);
    ttl += ttl_delta;
    ttl /= 10;

    i = 0;

#define EMIT(a, n) if(i > MAXSIZE - n) goto fail; memcpy(buf + i, a, n); i += n;

#define EMIT1(b) if(i > MAXSIZE - 1) goto fail; buf[i++] = b;

#define EMIT2(ss) do { \
    if(i > MAXSIZE - 2) goto fail; \
    twobytes.s = htonl(ss); \
    buf[i++] = twobytes.c[0]; \
    buf[i++] = twobytes.c[1]; } while(0)

#define EMIT4(n) do {               \
    if(i > MAXSIZE - 4) goto fail; \
    fourbytes.i = htonl(n); \
    buf[i++] = fourbytes.c[0]; \
    buf[i++] = fourbytes.c[1]; \
    buf[i++] = fourbytes.c[2]; \
    buf[i++] = fourbytes.c[3]; } while (0)

#define EMIT8(l) do { \
    if(i > MAXSIZE - 8) goto fail; \
    eightbytes.i = htonll(l); \
    buf[i++] = eightbytes.c[0]; \
    buf[i++] = eightbytes.c[1]; \
    buf[i++] = eightbytes.c[2]; \
    buf[i++] = eightbytes.c[3]; \
    buf[i++] = eightbytes.c[4]; \
    buf[i++] = eightbytes.c[5]; \
    buf[i++] = eightbytes.c[6]; \
    buf[i++] = eightbytes.c[7]; } while(0)

    EMIT1(PROTOCOL_VERSION);
    EMIT8(serial);
    EMIT4(ttl);

    if(network_name) {
        EMIT1(OPT_NETWORK_NAME);
        EMIT1(strlen(network_name) + 2);
        EMIT(network_name, strlen(network_name));
    }

    if(ipv6_prefix) {
        EMIT1(OPT_IPv6_PREFIX);
        EMIT1(18);
        EMIT(ipv6_prefix, 16);
    }

    if(routing_protocol == ROUTING_PROTOCOL_NONE) {
        EMIT1(OPT_ROUTING_PROTOCOL);
        EMIT1(3);
        EMIT1(ROUTING_PROTOCOL_NONE);
    } else if(routing_protocol == ROUTING_PROTOCOL_STATIC) {
        EMIT1(OPT_ROUTING_PROTOCOL);
        EMIT1(21);
        EMIT1(ROUTING_PROTOCOL_STATIC);
        EMIT1(STATIC_DEFAULT_GATEWAY);
        EMIT1(18);
        EMIT(default_gateway, 16);
    } else if(routing_protocol == ROUTING_PROTOCOL_OLSR) {
        EMIT1(OPT_ROUTING_PROTOCOL);
        EMIT1(21);
        EMIT1(ROUTING_PROTOCOL_OLSR);
        EMIT1(OLSR_MULTICAST_ADDRESS);
        EMIT1(18);
        EMIT(multicast_address, 16)
    }

    if(name_server) {
        EMIT1(OPT_NAME_SERVER);
        EMIT1(18);
        EMIT(name_server, 16);
    }

    data_length = i;

    EMIT1(OPT_SIGNATURE);
    EMIT1(9);
    EMIT1(SIGNATURE_UNSIGNED);
    EMIT2(data_length + 5);
    EMIT(toto, 4);

    write(1, buf, i);
    return 0;

 fail:
    fprintf(stderr, "Buffer overflow");
    exit(1);
}


